import contextlib
import sys
from argparse import Namespace
from io import StringIO
from typing import Generator

import pytest

from mitosis.__main__ import _create_parser
from mitosis.__main__ import _process_cl_args
from mitosis.__main__ import _split_param_str
from mitosis.__main__ import main
from mitosis.__main__ import normalize_modinput
from mitosis.tests.mock_legacy import lookup_dict
from mitosis.tests.mock_legacy import run
from mitosis.tests.mock_paper import meth_config
from mitosis.tests.mock_part1 import Klass


@pytest.mark.parametrize(
    ("params", "eval_params"),
    (([], []), (["foo=test"], []), ([], ["seed=1"])),
    ids=("no args", "lookup", "eval"),
)
def test_legacy_module(params, eval_params):
    args = Namespace(
        experiment=[],
        module="mitosis.tests.mock_legacy",
        debug=True,
        config="pyproject.toml",
        folder=None,
        eval_param=eval_params,
        param=params,
    )
    result = _process_cl_args(args)
    assert len(result["steps"]) == 1
    assert result["steps"][0].name == "mitosis.tests.mock_legacy"
    assert result["steps"][0].action == run
    assert id(result["steps"][0].lookup) == id(lookup_dict)


def test_experiment_arg():
    args = Namespace(
        experiment=["data", "fit_eval"],
        module=None,
        debug=True,
        config="mitosis/tests/test_pyproject.toml",
        folder=None,
        eval_param=["data.extra=True"],
        param=["data.length=test", "fit_eval.metric=test"],
    )
    result = _process_cl_args(args)
    assert len(result["steps"]) == 2
    assert [step.name for step in result["steps"]] == ["data", "fit_eval"]
    assert result["steps"][0].action == Klass.gen_data
    assert id(result["steps"][1].lookup) == id(meth_config)


def test_argparse_options():
    parser = _create_parser()
    args = parser.parse_args(
        ["-m", "mod", "-d", "--config", "foo.toml", "-F", "foo/bar"]
    )
    assert args.experiment == []
    assert args.module == "mod"
    assert args.debug is True
    assert args.config == "foo.toml"
    assert args.folder == "foo/bar"
    assert args.eval_param == []
    assert args.param == []


def test_argparse_main():
    parser = _create_parser()
    args = parser.parse_args(
        [
            "step1", "step2", "-e", "a=1", "-e", "b=2", "-p", "c=d", "-p", "e=f", "-p", "g=h"  # fmt: skip; # noqa: E501
        ]
    )
    assert args.experiment == ["step1", "step2"]
    assert args.module is None
    assert args.debug is False
    assert args.config == "pyproject.toml"
    assert args.folder is None
    assert len(args.eval_param) == 2
    assert len(args.param) == 3


def test_split_param_str():
    result = _split_param_str("+a=b")
    assert result == ("", False, "a", "b")
    result = _split_param_str("a.b=c")
    assert result == ("a", True, "b", "c")


def test_normalize_modinput():
    modinput = "mitosis.tests.mock_experiment"
    result = normalize_modinput(modinput)
    assert result == {
        "mitosis.tests.mock_experiment": (
            "mitosis.tests.mock_experiment:run",
            "mitosis.tests.mock_experiment:lookup_dict",
        )
    }
    # if modinput is an object, connect to run and lookup_dict with . not :
    modinput = "mitosis.tests.mock_experiment:MockExp.MockExpInner"
    result = normalize_modinput(modinput)
    assert result == {
        "mitosis.tests.mock_experiment:MockExp.MockExpInner": (
            "mitosis.tests.mock_experiment:MockExp.MockExpInner.run",
            "mitosis.tests.mock_experiment:MockExp.MockExpInner.lookup_dict",
        )
    }


def test_version():
    @contextlib.contextmanager
    def set_argv(*args: str) -> Generator[None, None, None]:
        temp = sys.argv
        try:
            sys.argv = list(args)
            yield
        finally:
            sys.argv = temp

    @contextlib.contextmanager
    def capture_stdout() -> Generator[StringIO, None, None]:
        stdout = sys.stdout
        new_out = StringIO()
        try:
            sys.stdout = new_out
            yield new_out
        finally:
            sys.stdout = stdout

    with set_argv("mitosis", "--version"):
        with capture_stdout() as out:
            main()
            result = out.getvalue()
    assert result.split()[0] == "mitosis"



from logging import getLogger


def run(**kwargs):
    getLogger(__name__).info("This is run every time")
    getLogger(__name__).debug("This is run in debug mode only")
    return {"main": 0}


lookup_dict = {"foo": {"test": 2}}



from functools import lru_cache
from pathlib import Path
from typing import Optional

import git
import toml


@lru_cache
def load_mitosis_steps(
    proj_file: Path | str = "pyproject.toml",
) -> dict[str, tuple[str, str]]:
    """Load experiment steps saved in pyproject.toml (or other config)

    Args:
        pyproj_file:
            Local or absolute path to pyproject file.  default is pyproject.toml
            if local path or default, use current git repo's top level directory
    Raises:
        Runtime error if cannot load steps, or if they are badly formed.
    """
    proj_file = _choose_toml(proj_file)
    with open(proj_file, "r") as f:
        config = toml.load(f)
    try:
        result = config["tool"]["mitosis"]["steps"]
    except KeyError:
        raise RuntimeError(
            f"{proj_file.absolute()} does not have a tools.mitosis.steps table"
        )
    if any(not isinstance(vals, list) or len(vals) != 2 for vals in result.values()):
        raise RuntimeError("tool.mitosis.steps table is malformed")
    return {k: tuple(v) for k, v in result.items()}


@lru_cache
def get_repo() -> git.Repo:
    repo = git.Repo(Path.cwd(), search_parent_directories=True)
    return repo


def _choose_toml(filename: Path | str | None) -> Path:
    repo = get_repo()
    directory = Path(repo.working_dir)
    if filename is None:
        return directory / "pyproject.toml"
    elif not Path(filename).is_absolute():
        return directory / filename
    return Path(filename)


def _locate_trial_folder(
    hexstr: str, *, trials_folder: Optional[Path | str] = None
) -> Path:
    if trials_folder is None:
        trials_folder = Path().absolute()
    else:
        trials_folder = Path(trials_folder).resolve()
    matches = trials_folder.glob(f"*{hexstr}")
    try:
        first = next(matches)
    except StopIteration:
        raise FileNotFoundError(f"Could not find a trial that matched {hexstr}")
    try:
        next(matches)
    except StopIteration:
        return first
    raise RuntimeError(f"Two or more matches found for {hexstr}")



from logging import getLogger

import numpy as np
from numpy.typing import NBitBase
from numpy.typing import NDArray


class Klass:
    @staticmethod
    def gen_data(
        length: int, extra: bool = False
    ) -> dict[str, NDArray[np.floating[NBitBase]] | bool]:
        getLogger(__name__).info("This is run every time")
        getLogger(__name__).debug("This is run in debug mode only")

        return {"data": np.ones(length, dtype=np.float_), "extra": extra}



from abc import ABCMeta
from dataclasses import dataclass
from dataclasses import field
from types import ModuleType
from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import ParamSpec


P = ParamSpec("P")
ExpRun = Callable[P, dict]


class Experiment(ModuleType, metaclass=ABCMeta):
    __name__: str
    __file__: str
    name: str
    lookup_dict: dict[str, dict[str, Any]]
    run: ExpRun


@dataclass
class Parameter:
    """An experimental parameter

    Arguments:
        var_name: short name for the variant (particular values) across use cases
        arg_name: name of arg known to experiment
        vals: value of the parameter
        eval: whether variant name should be evaluated or looked up.
    """

    var_name: str
    arg_name: str
    vals: Any
    # > 3.10 only: https://stackoverflow.com/a/49911616/534674
    evaluate: bool = field(default=False, kw_only=True)


class ExpStep(NamedTuple):
    name: str
    action: ExpRun
    action_ref: str
    lookup: dict[str, Any]
    lookup_ref: str
    group: str | None
    args: list[Parameter]
    untracked_args: list[str]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "mitosis"
copyright = "2024, Jake Stevens-Haas"
author = "Jake Stevens-Haas"
release = "0.3.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
# html_static_path = ["_static"]



import argparse
from itertools import groupby
from pathlib import Path
from typing import Any
from typing import NamedTuple

from . import __version__
from . import _disk
from . import run
from . import unpack
from ._typing import ExpRun
from ._typing import ExpStep
from ._typing import Parameter


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="mitosis",
        description=("Orchestrate experiment to ensure reproducibility."),
    )
    parser.add_argument(
        "experiment",
        nargs="*",
        help="Name to identify the experiment step(s), specified in pyproject.toml",
    )
    parser.add_argument(
        "-m",
        "--module",
        nargs="?",
        help="Load complete experiment from a module",
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="Print version information and exit",
    )
    parser.add_argument(
        "--debug",
        "-d",
        action="store_true",
        help=(
            "Run in debug mode, allowing one to use uncommitted code changes and not"
            " recording results"
        ),
    )
    parser.add_argument(
        "--config",
        type=str,
        default="pyproject.toml",
        help="Name or path of config file",
    )
    parser.add_argument(
        "--folder",
        "-F",
        type=str,
        default=None,
        help="Where to save trials, relative to the experiment module folder",
    )
    parser.add_argument(
        "--eval-param",
        "-e",
        type=str,
        default=[],
        action="append",
        help=(
            "Parameters directly passed on command line.  Make sure to quote if you"
            " want argument to evaluate as a string"
        ),
    )
    parser.add_argument(
        "--param",
        "-p",
        action="append",
        default=[],
        help=(
            "Name of parameters to use with this trial, in format 'key=value'\ne.g."
            "--param solver=solver_1\nKeys must be understood by the experiment"
            " being run.  Values reference variables\nstored by the same name in the"
            " experiment's config dict.\n\n To skip tracking and locking a parameter, "
            "precede the key with a plus sign ('+')."
        ),
    )
    return parser


class CLIParam(NamedTuple):
    step_key: str
    track: bool
    arg_name: str
    var_name: str


def _split_param_str(paramstr: str) -> CLIParam:
    arg_name, var_name = paramstr.split("=")
    track = True
    if arg_name[0] == "+":
        track = False
        arg_name = arg_name[1:]
    ex_step, _, arg_name = arg_name.rpartition(".")
    return CLIParam(ex_step, track, arg_name, var_name)


def _process_cl_args(args: argparse.Namespace) -> dict[str, Any]:
    ep_tups = [_split_param_str(epstr) for epstr in args.eval_param]
    lp_tups = [_split_param_str(lpstr) for lpstr in args.param]

    if args.experiment:
        if args.module:
            raise RuntimeError("Cannot use -m option if also passing experiment steps.")
        proj_steps = _disk.load_mitosis_steps(args.config)
        steps = dict(filter(lambda k: k[0] in args.experiment, proj_steps.items()))
    elif args.module:
        mod: str = args.module
        steps = normalize_modinput(args.module)
        ep_tups = [CLIParam(mod, track, name, val) for _, track, name, val in ep_tups]
        lp_tups = [CLIParam(mod, track, name, val) for _, track, name, val in lp_tups]
    else:
        raise RuntimeError(
            "Must set either pass a list of experiment steps "
            "(defined in pyproject.toml) or use the -m flag to pass a single"
            "installed experiment module"
        )
    all_steps: dict[str, tuple[tuple[str, ExpRun], tuple[str, dict[str, Any]]]] = {
        k: ((runner, unpack(runner)), (lookup, unpack(lookup)))
        for k, (runner, lookup) in steps.items()
    }

    def group_and_pop(
        list_of_clargs: list[CLIParam],
    ) -> dict[str, list[tuple[bool, str, str]]]:
        new_list = sorted(list_of_clargs, key=lambda tup: tup.step_key)
        arg_groups = {
            k: list(v) for k, v in groupby(new_list, key=lambda tup: tup.step_key)
        }
        popped = {k: [tup[1:] for tup in l] for k, l in arg_groups.items()}
        return popped

    ep_dict = group_and_pop(ep_tups)
    lp_dict = group_and_pop(lp_tups)

    unassigned = set(ep_dict.keys()).union(lp_dict.keys()) - set(all_steps.keys())
    if unassigned:
        raise RuntimeError(
            f"Steps {unassigned} not in experiment, but arguments assigned to them"
        )

    def create_step(
        name: str,
        part_step: tuple[tuple[str, ExpRun], tuple[str, dict[str, Any]]],
        eps: list[tuple[bool, str, str]],
        lps: list[tuple[bool, str, str]],
    ) -> ExpStep:
        (run_ref, runnable), (lookup_ref, lookup) = part_step
        params = []
        untracked_args = []
        for track, arg_name, var_name in eps:
            params.append(Parameter(var_name, arg_name, var_name, evaluate=True))
            if not track:
                untracked_args.append(arg_name)
        for track, arg_name, var_name in lps:
            params.append(Parameter(var_name, arg_name, lookup[arg_name][var_name]))
            if not track:
                untracked_args.append(arg_name)
        return ExpStep(
            name, runnable, run_ref, lookup, lookup_ref, None, params, untracked_args
        )

    # groups[step] could be empty...
    exp_steps = [
        create_step(
            step_name,
            all_steps[step_name],
            ep_dict.get(step_name, []),
            lp_dict.get(step_name, []),
        )
        for step_name in all_steps.keys()
    ]

    if args.folder is None:
        folder = Path(_disk.get_repo().working_dir) / "trials"
    else:
        folder = Path(args.folder)
    return {
        "steps": exp_steps,
        "debug": args.debug,
        "trials_folder": folder,
    }


def normalize_modinput(obj_ref: str) -> dict[str, tuple[str, str]]:
    modname, _, qualname = obj_ref.partition(":")
    if qualname:
        sep = ":" + qualname + "."
    else:
        sep = ":"
    return {obj_ref: (modname + sep + "run", modname + sep + "lookup_dict")}


def main() -> None:
    parser = _create_parser()
    args = parser.parse_args()
    if args.version:
        try:
            import setuptools_scm  # type: ignore

            vstring = setuptools_scm.get_version()
        except Exception:
            vstring = __version__
        print("mitosis", vstring)
        return
    kwargs = _process_cl_args(args)
    run(**kwargs)


if __name__ == "__main__":
    main()



import logging
import pprint
import sys
from collections import OrderedDict
from datetime import datetime
from datetime import timezone
from glob import glob
from importlib import import_module
from importlib.metadata import packages_distributions
from importlib.metadata import version
from logging import Logger
from pathlib import Path
from random import choices
from time import process_time
from types import BuiltinFunctionType
from types import BuiltinMethodType
from types import FunctionType
from types import MethodType
from typing import Any
from typing import cast
from typing import Collection
from typing import Hashable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence

import dill  # type: ignore
import nbformat
import pandas as pd
import sqlalchemy as sql
from nbclient.exceptions import CellExecutionError
from nbconvert.exporters import HTMLExporter
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.writers.files import FilesWriter
from nbformat import NotebookNode
from sqlalchemy import Column
from sqlalchemy import create_engine
from sqlalchemy import Float
from sqlalchemy import insert
from sqlalchemy import inspection
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import select
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import update

from . import _disk
from ._disk import _locate_trial_folder
from ._typing import ExpStep
from ._typing import Parameter
from ._version import version as __version__  # noqa: F401


def trials_columns():
    return [
        Column("variant", String, primary_key=True),
        Column("iteration", Integer, primary_key=True),
        Column("commit", String, nullable=False),
        Column("cpu_time", Float),
        Column("results", String),
        Column("filename", String),
    ]


def variant_types():
    return [
        Column("name", String, primary_key=True),
        Column("params", String, unique=False),
    ]


def load_trial_data(hexstr: str, *, trials_folder: Optional[Path | str] = None):
    trial = _locate_trial_folder(hexstr, trials_folder=trials_folder)
    all_files = glob("results*.dill", root_dir=trial)
    results = []
    for file in all_files:
        with open(trial / file, "rb") as fh:
            results.append(dill.load(fh))
    return results


def _lookup_param(
    arg_name: str, var_name: str, lookup_dict: dict[str, Any]
) -> Parameter:
    stored = lookup_dict[arg_name][var_name]
    if isinstance(stored, Parameter):
        return Parameter(var_name, arg_name, stored.vals, evaluate=False)
    else:
        return Parameter(var_name, arg_name, stored, evaluate=False)


class DBHandler(logging.Handler):
    def __init__(
        self,
        filename: Path | str,
        table_name: str,
        cols: List[Column],
        separator: str = "--",
    ):
        self.separator = separator
        if Path(filename).is_absolute():
            self.db = Path(filename)
        else:
            self.db = Path(__file__).resolve().parent / filename

        md = MetaData()
        self.log_table = Table(table_name, md, *cols)
        url = "sqlite:///" + str(self.db)
        self.eng = create_engine(url)
        with self.eng.connect() as conn:
            if not inspection.inspect(conn).has_table(table_name):
                md.create_all(conn)

        super().__init__()
        self.addFilter(lambda rec: self.separator in rec.getMessage())

    def emit(self, record: logging.LogRecord):
        vals = self.parse_record(record.getMessage())
        stmt: sql.Insert | sql.Update
        if "insert" in vals[0]:
            stmt = insert(self.log_table)
            for i, col in enumerate(self.log_table.columns):
                if vals[i + 1]:
                    stmt = stmt.values({col: vals[i + 1]})
        elif "update" in vals[0]:
            stmt = update(self.log_table)
            for i, col in enumerate(self.log_table.columns):
                if col.primary_key:
                    stmt = stmt.where(col == vals[i + 1])
                else:
                    if vals[i + 1]:
                        stmt = stmt.values({col: vals[i + 1]})
        else:
            raise ValueError("Cannot parse db message")
        with self.eng.connect() as conn:
            conn.execute(stmt)

    def parse_record(self, msg: str) -> List[str]:
        return msg.split(self.separator)


def _init_logger(trial_log: Path, table_name: str, debug: bool) -> tuple[Logger, Table]:
    """Create a Trials logger with a database handler"""
    exp_logger = logging.Logger("experiments")
    exp_logger.setLevel(20)
    exp_logger.addHandler(logging.StreamHandler())
    db_h = DBHandler(trial_log, table_name, trials_columns())
    if len(exp_logger.handlers) < 2 and not debug:  # A weird error requires this
        exp_logger.addHandler(db_h)
    return exp_logger, db_h.log_table


def _init_variant_table(trial_db: Path, step: str, param: Parameter) -> Table:
    eng = create_engine("sqlite:///" + str(trial_db))
    md = MetaData()
    var_table = Table(f"{step}_variant_{param.arg_name}", md, *variant_types())
    inspector = inspection.inspect(eng)
    if not inspector.has_table(f"{step}_variant_{param.arg_name}"):
        md.create_all(eng)
    return var_table


def _verify_variant_name(trial_db: Path, step: str, param: Parameter) -> None:
    """Check for conflicts between variant names in prior trials

    Side effects:
        - If trial_db does not exist, will create it
        - If variant name has not been used before, will insert it
    """
    eng = create_engine("sqlite:///" + str(trial_db))
    md = MetaData()
    tb = Table(f"{step}_variant_{param.arg_name}", md, *variant_types())
    vals: Collection[Any]
    if isinstance(param.vals, Mapping):
        vals = StrictlyReproduceableDict({k: v for k, v in sorted(param.vals.items())})
    elif isinstance(param.vals, Collection) and not isinstance(param.vals, str):
        try:
            vals = StrictlyReproduceableList(sorted(param.vals))
        except (ValueError, TypeError):
            vals = param.vals
    else:
        vals = param.vals
    df = pd.read_sql(select(tb), eng)
    ind_equal = df.loc[:, "name"] == param.var_name
    if ind_equal.sum() == 0:
        stmt = tb.insert().values({"name": param.var_name, "params": str(vals)})
        with eng.connect() as conn:
            conn.execute(stmt)
    elif df.loc[ind_equal, "params"].iloc[0] != str(vals):
        raise RuntimeError(
            f"Parameter '{param.arg_name}' variant '{param.var_name}' "
            f"is stored with different values in {trial_db}, table '{tb}'. "
            f"(Stored: {df.loc[ind_equal, 'params'].iloc[0]}), attmpeted: {str(vals)}."
        )
    # Otherwise, parameter has already been registered and no conflicts


def _id_variant_iteration(
    trial_db: Path, trials_table: Table, master_variant: str
) -> int:
    """Identify the iteration for this exact variant of the trial

    Args:
        trial_log (path-like): location of the trial log database
        trials_table (sqlalchemy.Table): the main record of each
            trial/variant
        var_table (sqlalchemy.Table): the lookup table for simulation
            variants
        sim_params (dict): parameters used in simulated experimental
            data
        id_table (sqlalchemy.Table): the lookup table for trial ids
        prob_params (dict): Parameters used to create the problem/solver
            in the experiment
    """
    eng = create_engine("sqlite:///" + str(trial_db))
    stmt = select(trials_table).where(trials_table.c.variant == master_variant)
    df = pd.read_sql(stmt, eng)
    if df.empty:
        return 1
    else:
        return df["iteration"].max() + 1


def _lock_in_variant(
    step: str,
    params: Sequence[Parameter],
    untracked_params: Collection[str],
    trial_db: Path,
    debug: bool,
) -> str:
    """Calculate the unique variant name combining all variants of parameters"""
    for param in params:
        if debug or param.arg_name in untracked_params:
            continue
        _init_variant_table(trial_db, step, param)
        _verify_variant_name(trial_db, step, param)
    var_names = [param.var_name for param in params]
    arg_names = [param.arg_name for param in params]
    if not arg_names:
        return "noparams"
    return f"{step}-" + "-".join(
        [x for _, x in sorted(zip(arg_names, var_names), key=lambda pair: pair[0])]
    )


def _get_commit_and_project_root(debug: bool) -> tuple[str, Path]:
    repo = _disk.get_repo()
    commit = "0000000" if debug else repo.head.commit.hexsha
    if not debug and repo.is_dirty():
        raise RuntimeError(
            "Git Repo is dirty.  For repeatable tests,"
            " clean the repo by committing or stashing all changes and "
            "untracked files."
        )
    return commit, Path(repo.working_dir)


def run(
    steps: list[ExpStep],
    debug: bool = False,
    *,
    trials_folder: Path,
    output_extension: str = "html",
    matplotlib_dpi: int = 72,
) -> str:
    """Run the selected experiment.

    Arguments:
        exps: The experiment steps to run
        debug (bool): Whether to run in debugging mode or not.
        trials_folder: The folder to store output, database, log, and metadata.
        output_extension: what output type to produce using nbconvert.
            Either 'html' or 'ipynb'.
        matplotlib_dpi: dpi for matplotlib images.  Not yet
            functional.

    Returns:
        The pseudorandom key to this experiment
    """

    commit, _ = _get_commit_and_project_root(debug)
    trials_folder = Path(trials_folder).absolute()
    if not trials_folder.exists():
        trials_folder.mkdir(parents=True)
    exp_name = "_".join(
        step.name + (f"_{step.group}" if step.group else "") for step in steps
    )
    dbfile = exp_name + ".db"
    trial_db = trials_folder / dbfile
    master_variant = "+".join(
        _lock_in_variant(step.name, step.args, step.untracked_args, trial_db, debug)
        for step in steps
    )
    experiments_table = f"trials_{exp_name}"
    for step in steps:
        if step.group is not None:
            step.args.append(
                Parameter(f"'{step.group}'", "group", step.group, evaluate=True)
            )
    exp_logger, trials_table = _init_logger(trial_db, experiments_table, debug)
    iteration = (
        0 if debug else _id_variant_iteration(trial_db, trials_table, master_variant)
    )
    rand_key = "".join(choices(list("0123456789abcde"), k=6))

    out_filename = _create_filename(
        master_variant, debug, iteration, rand_key, output_extension
    )
    exp_metadata_folder = _make_metadata_folder(trials_folder, rand_key)
    _write_freezefile(exp_metadata_folder)

    start_time = _log_start_experiment(
        exp_logger, master_variant, iteration, commit, debug
    )
    nb, metric, exc = _run_in_notebook(
        steps,
        exp_metadata_folder,
        matplotlib_dpi,
        debug=debug,
    )
    _save_notebook(nb, out_filename, trials_folder, output_extension)
    (exp_metadata_folder / "experiment").symlink_to(trials_folder / out_filename)
    _log_finish_experiment(
        exp_logger, master_variant, iteration, commit, metric, out_filename, start_time
    )

    if exc is not None:
        raise exc
    return rand_key


def _run_in_notebook(
    steps: list[ExpStep],
    trials_folder: Path,
    matplotlib_dpi=72,
    debug: bool = False,
) -> tuple[nbformat.NotebookNode, Optional[str], Optional[Exception]]:
    code = (
        "import logging\n"
        "from pathlib import Path\n\n"
        "import matplotlib as mpl\n"
        "import dill\n"
        "import mitosis\n\n"
        f"mpl.rcParams['figure.dpi'] = {matplotlib_dpi}\n"
        f"mpl.rcParams['savefig.dpi'] = {matplotlib_dpi}\n"
        "inputs = None\n"
    )
    nb = nbformat.v4.new_notebook()
    setup_cell = nbformat.v4.new_code_cell(source=code)
    step_loader_cells: list[NotebookNode] = []
    step_runner_cells: list[NotebookNode] = []
    for order, step in enumerate(steps):
        lookup_params = {a.arg_name: a.var_name for a in step.args if not a.evaluate}
        eval_params = {a.arg_name: a.var_name for a in step.args if a.evaluate}
        logfile = trials_folder / "experiment.log"
        code = (
            (
                f"step_{order} = mitosis.unpack('{step.action_ref}')\n"
                f"lookup_{order} = mitosis.unpack('{step.lookup_ref}')\n"
                f"resolved_args_{order} = {{}}\n"
                f"logger = logging.getLogger('{step.action.__module__}')\n"
            )
            + (
                "logger.setLevel(logging.DEBUG)\n"
                if debug
                else "logger.setLevel(logging.INFO)\n"
            )
            + (
                f"logger.addHandler(logging.FileHandler('{str(logfile)}', delay=True))\n"  # noqa E501
                f"logger.info('Initialized experiment logger')\n"
                f'print("Loaded step {order} as {step.action_ref}")\n'
                f'print("Loaded lookup {order} as {step.lookup_ref}")\n'
                f"for arg_name, var_name in {lookup_params}.items():\n"
                f"    val = mitosis._lookup_param(arg_name, var_name, lookup_{order}).vals\n"  # noqa E501
                f"    resolved_args_{order}.update({{arg_name: val}}) \n"
                f"    print(arg_name,'=',resolved_args_{order}[arg_name])\n\n"
                f"for arg_name, var_name in {eval_params}.items():\n"
                f"    val = eval(var_name)\n"
                f"    resolved_args_{order}.update({{arg_name: val}}) \n"
                f"    print(arg_name,'=',resolved_args_{order}[arg_name])\n\n"
                f"mitosis._prettyprint_config(Path('{trials_folder}'), resolved_args_{order})\n"  # noqa E501
                f"print('Saving metadata to {trials_folder}')\n"
            )
        )
        step_loader_cells.append(nbformat.v4.new_code_cell(source=code))

        code = (
            f"if inputs is not None:\n"
            f"    curr_result = step_{order}(inputs, **resolved_args_{order})\n"
            f"else:\n"
            f"    curr_result = step_{order}(**resolved_args_{order})\n"
            f"with open(r'{trials_folder / (f'results_{order}.dill')}', 'wb') as f:\n"  # noqa E501
            f"    dill.dump(curr_result, f)\n"
            f"print(repr(curr_result))\n"
            f"inputs = curr_result.get('data', None)\n"
        )
        step_runner_cells.append(nbformat.v4.new_code_cell(source=code))

    nb["cells"] = [setup_cell] + step_loader_cells + step_runner_cells
    with open(trials_folder / "source.py", "w") as fh:
        fh.write("".join(cell["source"] for cell in nb.cells))
    ep = ExecutePreprocessor(timeout=-1)

    exception = None
    metrics = None
    if debug:
        allowed = cast(tuple[type[Exception], ...], ())
    else:
        allowed = (CellExecutionError,)
    try:
        ep.preprocess(nb, {"metadata": {"path": trials_folder}})
        metrics = nb["cells"][-1]["outputs"][0]["text"][:-1]
    except allowed as exc:
        exception = exc
    return nb, metrics, exception


def _save_notebook(nb, filename, trials_folder, extension):
    if extension == "html":
        html_exporter = HTMLExporter({"template_file": "lab"})
        (body, resources) = html_exporter.from_notebook_node(nb)
        base_filename = filename[:-5]  # FilesWriter adds a .html
        file_writer = FilesWriter()
        file_writer.build_directory = str(trials_folder)
        file_writer.write(body, resources, notebook_name=base_filename)
    elif extension == "ipynb":
        with open(str(trials_folder / filename), "w", encoding="utf-8") as f:
            nbformat.write(nb, f)
    else:
        raise ValueError


def cleanstr(obj):
    if (
        isinstance(obj, FunctionType)
        or isinstance(obj, MethodType)
        or isinstance(obj, BuiltinFunctionType)
        or isinstance(obj, BuiltinMethodType)
    ):
        if obj.__name__ == "<lambda>":
            raise ValueError("Cannot use lambda functions in this context")
        import_error = ImportError(
            "Other modules must be able to import stored functions and modules:"
            f"function named {obj.__qualname__} stored in {obj.__module__}"
        )
        if obj.__module__ == "__main__":
            raise import_error
        if "<locals>" in obj.__qualname__:
            raise import_error
        try:
            mod = sys.modules[obj.__module__]
        except KeyError:
            raise import_error
        if not hasattr(mod, obj.__qualname__) or getattr(mod, obj.__qualname__) != obj:
            raise import_error
        return f"<{type(obj).__name__} {obj.__module__}.{obj.__qualname__}>"
    else:
        if isinstance(obj, str):
            return f"'{str(obj)}'"
        return str(obj)


class StrictlyReproduceableDict(OrderedDict):
    """A Dict that enforces reproduceable string representations

    The standard function.__str__ includes a memory location, which means the
    string representation of a function changes every time a program is run.
    In order to provide some stability in logs indicating that a function was
    run, the following dictionary will reject creating a string from functions
    that are difficult or impossible to reproduce in experiments.  It will also
    produce a reasonable string representation without the memory address for
    functions that are reproduceable.
    """

    def __str__(self):
        string = "{"
        for k, v in self.items():
            string += f"{cleanstr(k)}: "
            if isinstance(v, Mapping):
                string += str(StrictlyReproduceableDict(**v)) + ", "
            elif isinstance(v, Collection) and not isinstance(v, Hashable):
                string += str(StrictlyReproduceableList(v)) + ", "
            else:
                string += f"{cleanstr(v)}, "
        else:
            string = string[:-2] + "}"
        return string


class StrictlyReproduceableList(List):
    def __str__(self):
        string = "["
        for item in iter(self):
            if isinstance(item, Mapping):
                string += str(StrictlyReproduceableDict(**item)) + ", "
            elif isinstance(item, Collection) and not isinstance(item, Hashable):
                string += str(StrictlyReproduceableList(item)) + ", "
            else:
                string += f"{cleanstr(item)}, "
        else:
            string = string[:-2] + "]"
        return string


def _make_metadata_folder(trials_folder: Path, rand_key: str) -> Path:
    metadata_key = datetime.now().astimezone().strftime(r"%Y-%m-%d") + f"_{rand_key}"
    metadata_folder = trials_folder / metadata_key
    metadata_folder.mkdir()
    return metadata_folder


def _create_filename(
    variant: str,
    debug: bool,
    iteration: int,
    suffix: Optional[str],
    extension: str,
) -> str:
    new_filename = f"trial_{variant}_{iteration}_{suffix}"
    if debug:
        new_filename += "debug"
    elif extension == "html":
        new_filename += ".html"
    elif extension == "ipynb":
        new_filename += ".ipynb"
    return new_filename


def _log_start_experiment(
    exp_logger: logging.Logger,
    variant: str,
    iteration: int,
    commit: str,
    debug: bool,
) -> float:
    exp_logger.info(f"trial entry: insert--{variant}--{iteration}--{commit}--------")
    utc_now = datetime.now(timezone.utc)
    cpu_now = process_time()
    log_msg = (
        f"Running experiment {variant}, "
        f"iteration {iteration} at time: "
        + utc_now.strftime("%Y-%m-%d %H:%M:%S %Z")
        + f".  Current repo hash: {commit}"
    )
    if debug:
        log_msg += ".  In debugging mode."
    exp_logger.info(log_msg)
    return cpu_now


def _log_finish_experiment(
    exp_logger: logging.Logger,
    variant: str,
    iteration: int,
    commit: str,
    metric: Optional[str],
    filename: str,
    start_time: float,
) -> None:
    utc_now = datetime.now(timezone.utc)
    exp_logger.info(
        "Finished experiment at time: "
        + utc_now.strftime("%Y-%m-%d %H:%M:%S %Z")
        + f".  Results: {metric}"
    )
    exp_logger.info(
        "trial entry: update"
        + f"--{variant}"
        + f"--{iteration}"
        + f"--{commit}"
        + f"--{process_time() - start_time}"
        + f"--{metric}"
        + f"--{filename}"
    )


def _write_freezefile(folder: Path):
    installed = {pkg for pkgs in packages_distributions().values() for pkg in pkgs}
    req_str = f"# {sys.version}\n"
    req_str += "\n".join(f"{pkg}=='{version(pkg)}'" for pkg in installed)
    with open(folder / "requirements.txt", "w") as f:
        f.write(req_str)


def _prettyprint_config(folder: Path, params: Collection[Parameter]):
    pretty = pprint.pformat(params)
    with open(folder / "config.txt", "a") as f:
        f.write(pretty)
        f.write("\n")


def unpack(obj_ref: str) -> Any:
    modname, _, qualname = obj_ref.partition(":")
    obj = import_module(modname)
    for attr in qualname.split("."):
        obj = getattr(obj, attr)
    return obj
